/*
 * FreeUDFLibC - copyright (c) 1999 Gregory Deatz
 *
 *   date_functions.c
 *     Date and time routines:
 *       add_month - add some number of months to a date
 *       add_year - add some number of years to a date
 *       age_in_months - Between "date" and "ref_date", how many months?
 *       age_in_weeks - Between "date" and "ref_date", how many weeks?
 *       cdow_long - Convert the date to a "long" pretty version
 *       cdow_short - Convert the date to a "short" pretty version
 *       cmonth_long - What is the long name of the month?
 *       cmonth_short - What is the short name of the month?
 *       day_of_month - What is the day of the month?
 *       day_of_week - What is the day of the week?
 *       day_of_year - What is the day of the year?
 *       is_leap_year - Is it leap year?
 *       max_date - Compare two dates and return the maximum.
 *       min_date - Compare two dates and return the minimum.
 *       month - What is the month?
 *       quarter - What is the quarter?
 *       strip_date - Strip the date portion of the date
 *       strip_time - Strip the time portion of the date
 *       str_to_time - Given a time string (hh:mm:ss AM/PM) return date of it
 *       week_of_year - What is the week of the year?
 *       woy - Return the year and the week concatenated (199901)
 *       year - What is the year?
 *       year_of_year - Given a date, return the year as it would
 *                      be returned from the week_of_year function.
 */

#include <time.h>
#include "global.h"
#include "date_functions.h"

static ISC_QUAD * gen_ib_date PROTO_LIST ((struct tm *));
static void i_week_of_year PROTO_LIST ((ISC_QUAD *, int *, int *));

#define SUNDAY 0
#define MONDAY 1
#define TUESDAY 2
#define WEDNESDAY 3
#define THURSDAY 4
#define FRIDAY 5
#define SATURDAY 6

#define JANUARY 0
#define FEBRUARY 1
#define MARCH 2
#define APRIL 3
#define MAY 4
#define JUNE 5
#define JULY 6
#define AUGUST 7
#define SEPTEMBER 8
#define OCTOBER 9
#define NOVEMBER 10
#define DECEMBER 11

static char * LONGDOW[] = {
  "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday",
  "Saturday"
};

static char * SHORTDOW[] = {
  "Sun", "Mon", "Tue", "Wed", "Thu", "Fri",
  "Sat"
};

static char * LONGMONTH[] = {
  "January", "February", "March", "April", "May", "June",
  "July", "August", "September", "October", "November",
  "December"
};

static char * SHORTMONTH[] = {
  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

ISC_QUAD * add_month (ib_date, months_to_add)
     ISC_QUAD *ib_date;
     int *months_to_add;
{
  struct tm t;
  int tm_mon, tm_year;
  
  isc_decode_date (ib_date, &t);
  tm_mon = t.tm_mon + (*months_to_add % 12);
  tm_year = t.tm_year + (*months_to_add / 12);
  if ((*months_to_add > 0) && (tm_mon < t.tm_mon))
    tm_year++;
  else if ((*months_to_add < 0) && (tm_mon > t.tm_mon))
    tm_year--;

  /*
   * Now, check that the month day is correctly adjusted.
   * 30 days has Sep, Apr, Jun, Nov all the rest have 31 except Feb
   */
  if (tm_mon != FEBRUARY) {
    if (t.tm_mday > 30) 
      switch (tm_mon) {
      case APRIL:
      case JUNE:
      case SEPTEMBER:
      case NOVEMBER:
	t.tm_mday = 30;
      }
  }
  else if (t.tm_mday > 28)
    t.tm_mday = isleap(tm_year) ? 29 : 28;

  t.tm_mon = tm_mon;
  t.tm_year = tm_year;
  
  return (ISC_QUAD *) gen_ib_date(&t);
}

ISC_QUAD * add_year (ib_date, years_to_add)
     ISC_QUAD *ib_date;
     int *years_to_add;
{
  struct tm t;
  
  isc_decode_date(ib_date, &t);
  t.tm_year += *years_to_add;
  
  /*
   * If it's February, and the day of the month is 29, and
   * it's *not* a leap year, then adjust the day
   */
  if (t.tm_mon == FEBRUARY && t.tm_mday > 28 && !isleap(t.tm_year))
    t.tm_mday = 28;
  
  return (ISC_QUAD *) gen_ib_date(&t);
}

int age_in_months (ib_date, ib_ref_date)
     ISC_QUAD *ib_date;
     ISC_QUAD *ib_ref_date;
{
  struct tm t, t_ref;
  
  isc_decode_date(ib_date, &t);
  isc_decode_date(ib_ref_date, &t_ref);
  
  return ((t.tm_year - t_ref.tm_year) * 12) + t.tm_mon - t_ref.tm_mon;
}

/*
 * Bases it's computation on the fact that they year has
 * an *average* of 52 weeks. So this is most certainly an
 * approximate function, but I'm quite confident that it is
 * a reasonably precise function.
 */
int age_in_weeks (ib_date, ib_ref_date)
     ISC_QUAD *ib_date;
     ISC_QUAD *ib_ref_date;
{
  struct tm t, t_ref;
  
  isc_decode_date(ib_date, &t);
  isc_decode_date(ib_ref_date, &t_ref);
  
  return
    ((t.tm_year - t_ref.tm_year) * 52) +
    (((t.tm_yday - t.tm_wday) -
      (t_ref.tm_yday - t_ref.tm_wday)) / 7);
}

char * cdow_long (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  char * result = (char *) MALLOC (LOC_STRING_SIZE);
  
  isc_decode_date(ib_date, &t);

  sprintf(result, "%s", LONGDOW[t.tm_wday]);
  
  return (char *) result;
}

char * cdow_short (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  char * result = (char *) MALLOC (LOC_STRING_SIZE);
  
  isc_decode_date(ib_date, &t);

  sprintf(result, "%s", SHORTDOW[t.tm_wday]);
  
  return (char *) result;
}

char * cmonth_long (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  char * result = (char *) MALLOC (LOC_STRING_SIZE);
  
  isc_decode_date(ib_date, &t);

  sprintf(result, "%s", LONGMONTH[t.tm_mon]);
  
  return (char *) result;
}

char * cmonth_short (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  char * result = (char *) MALLOC (LOC_STRING_SIZE);
  
  isc_decode_date(ib_date, &t);

  sprintf(result, "%s", SHORTMONTH[t.tm_mon]);
  
  return (char *) result;
}

int day_of_month (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  return t.tm_mday;
}

int day_of_week (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  return t.tm_wday + 1;
}

int day_of_year (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  return t.tm_yday + 1;
}

int is_leap_year (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  return isleap(t.tm_year);
}

ISC_QUAD * max_date (ib_date_1, ib_date_2)
     ISC_QUAD *ib_date_1;
     ISC_QUAD *ib_date_2;
{
  ISC_QUAD * result = (ISC_QUAD *) MALLOC(sizeof(ISC_QUAD));
  
  if ((ib_date_1->isc_quad_high > ib_date_2->isc_quad_high) ||
      ((ib_date_1->isc_quad_high == ib_date_2->isc_quad_high) &&
       (ib_date_1->isc_quad_low > ib_date_2->isc_quad_low)))
    *result = *ib_date_1;
  else
    *result = *ib_date_2;

  return result;
}

ISC_QUAD * min_date (ib_date_1, ib_date_2)
     ISC_QUAD *ib_date_1;
     ISC_QUAD *ib_date_2;
{
  ISC_QUAD * result = (ISC_QUAD *) MALLOC(sizeof(ISC_QUAD));
  
  if ((ib_date_1->isc_quad_high < ib_date_2->isc_quad_high) ||
      ((ib_date_1->isc_quad_high == ib_date_2->isc_quad_high) &&
       (ib_date_1->isc_quad_low < ib_date_2->isc_quad_low)))
    *result = *ib_date_1;
  else
    *result = *ib_date_2;

  return result;
}

int month (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  return t.tm_mon + 1;
}

int quarter (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  return (t.tm_mon / 4) + 1;
}

ISC_QUAD * strip_date (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  t.tm_mday = 0;
  t.tm_mon = 0;
  t.tm_year = 0;
  t.tm_wday = 0;
  t.tm_yday = 0;

  return gen_ib_date(&t);
}

ISC_QUAD * strip_time (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  t.tm_sec = 0;
  t.tm_min = 0;
  t.tm_hour = 0;
  t.tm_isdst = 0;

  return gen_ib_date(&t);
}

ISC_QUAD * str_to_time (sz)
     char *sz;
{
  struct tm t;
  int len = strlen(sz);
  int h, m, s, am;
  ISC_QUAD * result = (ISC_QUAD *) MALLOC(sizeof(ISC_QUAD));

  h = strtol(sz, NULL, 10);
  m = len > 3 ? strtol(&sz[3], NULL, 10) : 0;
  s = len > 6 ? strtol(&sz[6], NULL, 10) : 0;
  am = len > 9 ? strcmp(&sz[9], "PM") : 0;
  
  t.tm_sec = s;
  t.tm_min = m;
  if (am)
    t.tm_hour = (h + ((h == 12) ? 12 : 0)) % 24;
  else
    t.tm_hour = (h + ((h == 12) ? 0 : 12)) % 24;
  t.tm_mday = 0;
  t.tm_mon = 0;
  t.tm_year = 0;
  t.tm_wday = 0;
  t.tm_yday = 0;
  t.tm_isdst = 0;

  isc_encode_date(&t, result);
  return result;
}

int week_of_year (ib_date)
     ISC_QUAD *ib_date;
{
  int week, year;

  i_week_of_year(ib_date, &week, &year);
  return week;
}

char * woy (ib_date)
     ISC_QUAD *ib_date;
{
  char * result = (char *) MALLOC (LOC_STRING_SIZE + 1);
  int week, year;

  i_week_of_year(ib_date, &week, &year);
  
  sprintf(result, "%d%2d", year, week);
  return result;
}

int year (ib_date)
     ISC_QUAD *ib_date;
{
  struct tm t;
  isc_decode_date(ib_date, &t);

  return t.tm_year + 1900;
}

int year_of_year (ib_date)
     ISC_QUAD *ib_date;
{
  int week, year;

  i_week_of_year(ib_date, &week, &year);
  return year;
}

static ISC_QUAD * gen_ib_date (t)
     struct tm *t;
{
  ISC_QUAD *ibd = (ISC_QUAD *) MALLOC (sizeof(ISC_QUAD));
  isc_encode_date(t, ibd);
  return ibd;
}


/*
 * Simple observations:
 *   7 * 52 = 364 = a common year + 1 (or 2 if it is a leap year)
 *   There are 365-366 days in the year.
 *   A week is Sunday - Saturday.
 * So....
 *
 * if the year begins on a Sun - Wed, then Jan 1 is considered to begin
 *   Week #1.
 * if the year begins on Thu - Sat, then Jan 1 is considered to still
 *   be part of the last week of the previous year.
 * if the year begins on a Wed OR
 *    the year begins on a Tue or Wed and its a leap year THEN
 *   the year has 53 weeks in it...
 */
static void i_week_of_year (ib_date, week, year)
     ISC_QUAD * ib_date;
     int * week;
     int * year;
{
  struct tm t;
  int dow_ybeg, yr, wk;
  
  isc_decode_date(ib_date, &t);

  yr = t.tm_year + 1900;

  /* When did the year begin? */
  dow_ybeg = (t.tm_yday < t.tm_wday)
    ? (t.tm_wday - t.tm_yday) % 7
    : (t.tm_yday - t.tm_wday) % 7;

  /* Get the Sunday beginning this week. */
  wk = (t.tm_yday - t.tm_wday);
  /*
   * If the Sunday beginning this week was last year, then
   *   if this year begins on a Wednesday or previous, then
   *   this is most certainly the first week of the year.
   *   if this year begins on a thursday or
   *     last year was a leap year and this year begins on a friday, then
   *   this week is 53 of last year.
   *   Otherwise this week is 52 of last year.
   */
  if (wk < 0) {
    if (dow_ybeg <= WEDNESDAY)
      wk = 1;
    else if ((dow_ybeg == THURSDAY) ||
	     (isleap((yr - 1)) && (dow_ybeg == FRIDAY))) {
      wk = 53;
      yr--;
    } else {
      wk = 52;
      yr--;
    }
  }
  else {
    /* If the Sunday beginning this week falls in this year
     *   if the year begins on a Sun, Mon, Tue or Wed then
     *     This week # is (Week + 7) div 7
     *   otherwise this week is
     *     Week div 7 + 1.
     *   if the week is > 52 then
     *     if this year began on a wed or this year is leap year and it
     *       began on a tuesday, then set the week to 53.
     *     otherwise set the week to 1 of *next* year.
     */
    if (dow_ybeg <= WEDNESDAY)
      wk = (wk + 7 + dow_ybeg) / 7;
    else
      wk = (wk / 7) + 1;
    if (wk > 52) {
      if ((dow_ybeg == WEDNESDAY) ||
	  (isleap(yr) && (dow_ybeg == TUESDAY)))
	wk = 53;
      else {
        wk = 1;
	yr++;
      }
    }
  }
  *week = wk;
  *year = yr;
}

